#ifndef ALAMO_H
#define ALAMO_H

#include <stdio.h>
#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
  size_t (*clen)(const char *p, size_t i, size_t l);
  unsigned long (*code)(const char *p, size_t i, size_t l);
  int (*save)(char *buff, unsigned long ch);
  char *re; /* source string */
  int rl; /* source string length */
  int en; /* error number */
  int ev; /* error information value */
  int nt; /* n of tokens */
  int ns; /* n of states */
  int no; /* n of opecodes */
  unsigned long *oe; /* opecode entries */
  unsigned long *op; /* opecode buffer */
  unsigned long *mk; /* marker (for makedot) */
  unsigned long *tl; /* last pos (top level) */
  unsigned long *tf; /* flag:^ */
  unsigned long *bf; /* flag:$ */
  short *tk; /* token buffer */
  short *tn; /* lookup token by state */
  short *fl; /* follow pos */
} ALMEXP;

enum {
  ALMLEXERR = 1, /* lexical */
  ALMSYNERR = 2, /* syntax */
  ALMMEMERR = 3, /* malloc */
  ALMEPSERR = 4, /* empty */
  ALMSTTERR = 5, /* operator only, no status */
  ALMSTKERR = 6, /* parser stack overflow */
  ALMRUNERR = 7, /* bug from runtime? */
  ALMPCLERR = 8, /* unsupported POSIX class */
};

enum { 
  ALMOPEND,
  ALMOPBLK,
  ALMOPNEG,
  ALMOPDOT,
  ALMOPCMP,
  ALMOPSET,
  ALMOPRNG,
  ALMOPFNC,
};

enum {
  ALMPARSNON,
  ALMPARSACC,
  ALMPARSSYN,
  ALMPARSSTK,
};

#define ALMENCMSJ "shift_jis"
#define ALMENCEUC "euc-jp"
#define ALMENCUTF "utf-8"

#define ALMBMLEN(nbit) (((nbit) + 31) / 32 * sizeof (unsigned long))
#define ALMSETBIT(m, n) ((m)[(n) / 32] |= (1 << ((n) % 32)))
#define ALMCLRBIT(m, n) ((m)[(n) / 32] &= ~(1 << ((n) % 32)))
#define ALMCHKBIT(m, n) ((m)[(n) / 32] & (1 << ((n) % 32)))
#define ALMBMADD(nb, dst, src) \
 { int i; for (i = 0; i < ALMBMLEN(nb); i++) (dst)[i] |= (src)[i]; }
#define ALMBMCPY(nb, dst, src) \
  memcpy(dst, src, ALMBMLEN(nb))
#define ALMBMCLR(nb, dst) \
  memset(dst, 0, ALMBMLEN(nb))

#define ALMFOLLOWPOS(p, s, d) (((p)->fl)[(s) * ((p)->ns) + (d)])
#define ALMTKENT 5
#define ALMTSTR(p, n) (&(((p)->re)[ALMTS((p), (n))]))
#define ALMLINE(p, n) ALMLN(p, ALMTN(p, n))
#define ALMOP(p, n) (&(((p)->op)[ALMOE((p), (n))]))
#define ALMTV(p, n) (((p)->tk)[((n) * ALMTKENT) + 0])
#define ALMTS(p, n) (((p)->tk)[((n) * ALMTKENT) + 1])
#define ALMTL(p, n) (((p)->tk)[((n) * ALMTKENT) + 2])
#define ALMSN(p, n) (((p)->tk)[((n) * ALMTKENT) + 3])
#define ALMLN(p, n) (((p)->tk)[((n) * ALMTKENT) + 4])
#define ALMTN(p, n) (((p)->tn)[n])
#define ALMOE(p, n) (((p)->oe)[n])
#define ALMACC(p, s) ALMCHKBIT((p)->tl, s)

ALMEXP *almcompile(const char *s, int len, const char *charset);
int almopeval(const char *s, int m, int b);
const char *almopesym(int k);
void almmakedot(FILE *fp, ALMEXP *e);
void almerror(FILE *fp, ALMEXP *e);
void almopdump(FILE *fp, ALMEXP *p, int stt);
void almopdumpall(FILE *fp, ALMEXP *p);
int almexeopc(ALMEXP *e, unsigned long *p, unsigned long c);
int almexecmp(ALMEXP *e, int n, unsigned long c);

#ifdef __cplusplus
}
#endif

#endif
