#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <vector>
#include <string>
#include <map>
#include <stdexcept>
#include <algorithm>

#include "alamo.h"

#define DUMPVM(e, addr) { \
  printf("// DEBUG:VM = {"); \
  for (int i = 0; i < e->ns; i++)  { \
    if (ALMCHKBIT(addr, i)) printf("S%d,", i); \
  } \
  printf("};\n"); \
}

class almvm {
  std::vector<unsigned long> cur, nxt;
  std::map<std::vector<unsigned long>, std::vector<unsigned long> > cache;
  int cb;
public:
  ALMEXP *e;
  almvm(const std::string &expr, const char *enc = ALMENCMSJ) {
    if (!(e = almcompile(expr.c_str(), expr.size(), enc))) {
      throw std::runtime_error("almcompile returns null");
    } else if (e->en) {
      throw e;
    }
    cb = ALMBMLEN(e->ns);
    cur.resize(cb + 1, 0);
    nxt.resize(cb);
    reset();
  }
  ~almvm() {
    free(e);
  }
  void reset() {
    std::fill(cur.begin(), cur.end(), 0);
    ALMSETBIT(cur, 0); /* S0 */
  }
  void step(unsigned long ch) { // deep copy
    std::map<std::vector<unsigned long>, 
     std::vector<unsigned long> >::iterator it;
    cur[cur.size() - 1] = ch;
    if ((it = cache.find(cur)) != cache.end()) {
      printf("// DEBUG:cache hit!\n");
      std::copy(it->second.begin(), it->second.end(), cur.begin());
      return;
    }
    std::fill(nxt.begin(), nxt.end(), 0);
    for (int s = 0; s < e->ns; s++) {
      if (ALMCHKBIT(&(cur[0]), s)) {
        for (int d = 1; d < e->ns; d++) {
	  if (ALMFOLLOWPOS(e, s, d) != -1) {
	    if (almexecmp(e, d, ch)) {
	      printf("// DEBUG:%d->%d\n", s, d);
	      ALMSETBIT(&(nxt[0]), d);
	      // DUMPVM(e, &(nxt[0]));
            }
	  }
	}
      }
    }
    cache[cur] = nxt;
    std::copy(nxt.begin(), nxt.end(), cur.begin());
    printf("// DEBUG:cache size=%d\n", cache.size());
  }
  int stat() {
    DUMPVM(e, &(cur[0]));
    int v = 0;
    for (int s = 1; s < e->ns; s++) {
      if (ALMCHKBIT(&(cur[0]), s)) {
        if (ALMACC(e, s)) {
          printf("// DEBUG:stat=acc\n");
	  return +1;
	}
        v++;
      }
    }
    printf("// DEBUG:stat=%d\n", v ? 0 : -1);
    return v ? 0 : -1;
  }
  bool matchonly(const char *s, size_t l) {
    int v;
    reset();
    for (size_t i = 0; cb = (*(e->clen))(s, i, l); i += cb) {
      step((*(e->code))(s, i, l));
      if (v = stat()) break;
    }
    return v == 1;
  }
  int match(const char *s, size_t l) {
    int max = 0;
    int v;
    reset();
    for (size_t i = 0; cb = (*(e->clen))(s, i, l); i += cb) {
      step((*(e->code))(s, i, l));
      if ((v = stat()) < 0) break;
      if (v == 1) max = (i + cb);
    }
    return max;
  }
};

int main(int argc, char *argv[]) {
  const char *reg;
  const char *str;
  switch (argc) {
  case 3:
    reg = argv[1];
    str = argv[2];
    break;
  default:
    fprintf(stderr, "usage:%s expr string\n", argv[0]);
    exit(1);
  }
  try {
    almvm vm(reg);
    almopdumpall(stdout, vm.e);
    almmakedot(stdout, vm.e);
    printf("match len = %d\n", vm.match(str, strlen(str)));
  } catch (std::exception &e) {
    fprintf(stderr, "%s:%s\n", reg, e.what()); 
    exit(-1);
  } catch (ALMEXP *e) {
    almerror(stderr, e);
    exit(-1);
  }
  exit(0);
}
