%include {
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include "sx.h" /* INCLUDE */

/*
%stack_size܂YYSTACKDEPTH̃AP[g
*/

#if defined(__BORLANDC__)
#pragma warn -8057 
#pragma warn -8019
#endif

typedef struct {
  ALMEXP *e;
  size_t depth;
  struct node *node;
  char *pool;
  int cc;
} ALMTMP;

struct node {
  char use;
  char na;
  unsigned long *fpos;
  unsigned long *lpos;
};

struct node *newnode(ALMTMP *ap) {
  int i;
  for (i = 0; i < ap->depth; i++) {
    if (!ap->node[i].use) {
      size_t bs;
      ap->node[i].use = 1;
      ap->node[i].na = 0;
      bs = ALMBMLEN(ap->e->ns);
      memset(ap->node[i].fpos, 0, bs + bs);
      return &(ap->node[i]);
    }
  }
  /* no reach, check YYSTACKDEPTH */
  printf("FATAL\n");
  exit(-1);
}

ALMTMP *makealmtmp(ALMEXP *e, int depth) { // YYSTACKDEPTH
  ALMTMP *p;
  size_t ns = sizeof (struct node);
  size_t bs = (size_t)ALMBMLEN(e->ns);
  size_t nsize = sizeof (struct node) * depth;
  size_t psize = (bs + bs) * depth;
  size_t total = sizeof (ALMTMP) + nsize + psize;
  if (p = malloc(total)) {
    int i;
    p->e = e;
    p->depth = depth;
    p->node = (struct node *)&(p[1]);
    p->pool = (char *)&(p->node[depth]);
    for (i = 0; i < depth; i++) {
      p->node[i].use = 0;
      p->node[i].fpos = (unsigned long *)&(p->pool[(bs + bs) * i]);
      p->node[i].lpos = (unsigned long *)&(p->pool[(bs + bs) * i + bs]);
    }
    p->cc = ALMPARSNON;
  }
  return p;
}

}

%name alamo

%type txt { struct node * }
%type str { struct node * }
%type lor { struct node * }
%type seq { struct node * }
%type exp { struct node * }
%type val { struct node * }

%ifdef DEBUG
%destructor txt { $$->use = 0; printf("DEBUG:freenode:%x\n", $$); }
%destructor str { $$->use = 0; printf("DEBUG:freenode:%x\n", $$); }
%destructor lor { $$->use = 0; printf("DEBUG:freenode:%x\n", $$); }
%destructor seq { $$->use = 0; printf("DEBUG:freenode:%x\n", $$); }
%destructor exp { $$->use = 0; printf("DEBUG:freenode:%x\n", $$); }
%destructor val { $$->use = 0; printf("DEBUG:freenode:%x\n", $$); }
%endif
%ifndef DEBUG
%destructor txt { $$->use = 0; }
%destructor str { $$->use = 0; }
%destructor lor { $$->use = 0; }
%destructor seq { $$->use = 0; }
%destructor exp { $$->use = 0; }
%destructor val { $$->use = 0; }
%endif

%token_type { unsigned short }

%extra_argument { ALMTMP *ap }
/* %token_prefix TK_ */

/* TOKENƂ}NtokensQƂł */

%parse_accept { ap->cc = ALMPARSACC; }
%syntax_error { ap->cc = ALMPARSSYN; }
%stack_overflow { ap->cc = ALMPARSSTK; }

%right AST PLS QST.
%left LOR.
%nonassoc EOE.

program ::= txt(A) EOE.
  {
    int x;
    for (x = 0; x < ap->e->ns; x++) {
      if (ALMCHKBIT(A->fpos, x)) {
        ALMFOLLOWPOS(ap->e, 0, x) = ALMTN(ap->e, x);
      }
    }
    ALMBMCPY(ap->e->ns, ap->e->tl, A->lpos);
  }

txt(A) ::= txt(B) RET str(C).
  { 
    A = newnode(ap); 
    A->na = (B->na | C->na);
    ALMBMCPY(ap->e->ns, A->fpos, B->fpos);
    ALMBMADD(ap->e->ns, A->fpos, C->fpos);
    ALMBMCPY(ap->e->ns, A->lpos, B->lpos);
    ALMBMADD(ap->e->ns, A->lpos, C->lpos);
  }

txt(A) ::= str(B).
  { A = B; }

str(A) ::= TOP lor(B) BOT.
  { int x;
    A = B;
    for (x = 0; x < ap->e->ns; x++) {
      if (ALMCHKBIT(B->fpos, x)) ALMSETBIT(ap->e->tf, x);
    }
    for (x = 0; x < ap->e->ns; x++) {
      if (ALMCHKBIT(B->lpos, x)) ALMSETBIT(ap->e->bf, x);
    }
  }

str(A) ::= TOP lor(B).
  { int x;
    A = B;
    for (x = 0; x < ap->e->ns; x++) {
      if (ALMCHKBIT(B->fpos, x)) ALMSETBIT(ap->e->tf, x);
    }
  }

str(A) ::= lor(B) BOT.
  { int x;
    A = B;
    for (x = 0; x < ap->e->ns; x++) {
      if (ALMCHKBIT(B->lpos, x)) ALMSETBIT(ap->e->bf, x);
    }
  }

str(A) ::= lor(B).
  { A = B; }

lor(A) ::= lor(B) LOR exp(C). 
  { 
    A = newnode(ap); 
    A->na = (B->na | C->na);
    ALMBMCPY(ap->e->ns, A->fpos, B->fpos);
    ALMBMADD(ap->e->ns, A->fpos, C->fpos);
    ALMBMCPY(ap->e->ns, A->lpos, B->lpos);
    ALMBMADD(ap->e->ns, A->lpos, C->lpos);
  }

lor(A) ::= exp(B).
  { A = B; }

exp(A) ::= exp(B) val(C).
  { int y, x;
    A = newnode(ap); 
    A->na = (B->na & C->na);
    ALMBMCPY(ap->e->ns, A->fpos, B->fpos);
    if (B->na) { ALMBMADD(ap->e->ns, A->fpos, C->fpos); }
    ALMBMCPY(ap->e->ns, A->lpos, C->lpos);
    if (C->na) { ALMBMADD(ap->e->ns, A->lpos, B->lpos); }
    for (y = 0; y < ap->e->ns; y++) {
      if (ALMCHKBIT(B->lpos, y)) {
        for (x = 0; x < ap->e->ns; x++) {
          if (ALMCHKBIT(C->fpos, x)) {
            ALMFOLLOWPOS(ap->e, y, x) = ALMTN(ap->e, x);
          }
        }
      }
    }
  }

exp(A) ::= val(B).
  { 
    A = B; 
  }

val(A) ::= val(B) AST. 
  { int y, x;
    A = newnode(ap); 
    A->na = 1;
    ALMBMCPY(ap->e->ns, A->fpos, B->fpos);
    ALMBMCPY(ap->e->ns, A->lpos, B->lpos);
    for (y = 0; y < ap->e->ns; y++) {
      if (ALMCHKBIT(A->lpos, y)) {
        for (x = 0; x < ap->e->ns; x++) {
          if (ALMCHKBIT(A->fpos, x)) {
            ALMFOLLOWPOS(ap->e, y, x) = ALMTN(ap->e, x);
          }
        }
      }
    }
  }

val(A) ::= val(B) PLS. 
  { int y, x;
    A = newnode(ap); 
    A->na = 0;
    ALMBMCPY(ap->e->ns, A->fpos, B->fpos);
    ALMBMCPY(ap->e->ns, A->lpos, B->lpos); 
    for (y = 0; y < ap->e->ns; y++) {
      if (ALMCHKBIT(A->lpos, y)) {
        for (x = 0; x < ap->e->ns; x++) {
          if (ALMCHKBIT(A->fpos, x)) {
            ALMFOLLOWPOS(ap->e, y, x) = ALMTN(ap->e, x);
          }
        }
      }
    }
  }

val(A) ::= val(B) QST. 
  { int y, x;
    A = newnode(ap); 
    A->na = 1;
    ALMBMCPY(ap->e->ns, A->fpos, B->fpos);
    ALMBMCPY(ap->e->ns, A->lpos, B->lpos); 
  }

val(A) ::= GRO lor(B) GRC.
  { A = B; }

val(A) ::= CHR(B). 
  { 
    A = newnode(ap);
    A->na = 0;
    ALMSETBIT(A->fpos, ALMSN(ap->e, B));
    ALMSETBIT(A->lpos, ALMSN(ap->e, B));
  }

val(A) ::= DOT(B).
  { 
    A = newnode(ap);
    A->na = 0;
    ALMSETBIT(A->fpos, ALMSN(ap->e, B));
    ALMSETBIT(A->lpos, ALMSN(ap->e, B));
  }
    
val(A) ::= SET(B). 
  { 
    A = newnode(ap);
    A->na = 0;
    ALMSETBIT(A->fpos, ALMSN(ap->e, B));
    ALMSETBIT(A->lpos, ALMSN(ap->e, B));
  }

