#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <string.h>
#include "mbs.h"
#include "alamo.h"
#include "tkst.h" /* INCLUDE */
#include "sx.c" /* INCLUDE */

static const char *escstr[] = { 
 "\\\\", "\\a", "\\b", "\\f", "\\n", "\\r", "\\t", "\\v", 
};
static const unsigned long escchr[] = { 
 '\\', '\a', '\b', '\f', '\n', '\r', '\t', '\v',
};

static const int nesc = (sizeof escstr) / (sizeof (escstr[0]));

static const char *opestr[] = 
 { "\n", "^", "$", ".", "?", "+", "*", "(", ")", "|", };
static const short opeval[] = 
 {  RET, TOP, BOT, DOT, QST, PLS, AST, GRO, GRC, LOR, };
static const int nops = (sizeof opestr) / (sizeof (opestr[0]));

static unsigned long crngdigit[] = { '0', '9', };
static unsigned long crngxdigit[] = { '0', '9', 'A', 'F', 'a', 'f', };
static unsigned long crngupper[] = { 'A', 'Z', };
static unsigned long crnglower[] = { 'a', 'z', };
static unsigned long crngalpha[] = { 'A', 'Z', 'a', 'z', };
static unsigned long crngalnum[] = { '0', '9', 'A', 'Z', 'a', 'z', };
static unsigned long csetspace[] = { ' ', '\t', '\n', 0x0b, '\f', '\r', };
static unsigned long csetblank[] = { ' ', '\t', };
static unsigned long crngascii[] = { 0x00, 0x7f, };
static unsigned long csetcntl[] = { 0x7f, };
static unsigned long crngcntl[] = { 0x00, 0x1f, };
static unsigned long crngpunct[] = { '!', '/', ':', '@', '[', '`', '{', '~', };
static unsigned long crnggraph[] = { '!', '~', };
static unsigned long crngprint[] = { '!', '~', };

static struct {
  const char *sym;
  int sl;
  int rl;
  unsigned long *set;
  unsigned long *rng;
} pclass[] = {
 { "[:digit:]", 0, 2, NULL, crngdigit },
 { "[:xdigit:]", 0, 2, NULL, crngxdigit },
 { "[:lower:]", 0, 2, NULL, crnglower },
 { "[:upper:]", 0, 2, NULL, crngupper },
 { "[:alpha:]", 0, 4, NULL, crngalpha },
 { "[:alnum:]", 0, 6, NULL, crngalnum },
 { "[:space:]", 6, 0, NULL, csetspace },
 { "[:blank:]", 2, 0, NULL, csetblank },
 { "[:ascii:]", 0, 2, NULL, crngascii },
 { "[:cntl:]", 1, 2, csetcntl, crngcntl },
 { "[:punct:]", 0, 8, NULL, crngpunct },
 { "[:graph:]", 0, 2, NULL, crnggraph },
 { "[:print:]", 0, 2, NULL, crngprint },
};
static const int npclass = (sizeof pclass) / (sizeof (pclass[0]));

#define MBSCMP(s, l, k) \
 ((((size_t)l) == strlen(k)) ? (memcmp(s, k, l) == 0) : 0)

#define CODEDUMP(out, p, v) { char buff[8]; \
  if (((v) < 0x20) || ((v) == 0x7f)) { \
    fprintf((out), "\\x%02x", v); \
  } else if ((0x20 <= (v)) && ((v) <= 0x7e)) { \
    fprintf((out), "%c", v); \
  } else { \
    (*((p)->save))(buff, (v)); \
    fprintf((out), "%s", buff); \
  } \
}

int almopeval(const char *s, int m, int b) {
  int i;
  for (i = 0; i < nops; i++) {
    if (MBSCMP(&(s[m]), b, opestr[i])) {
      return opeval[i];
    }
  }
  return 0;
}

const char *almopesym(int k) {
  int i;
  switch (k) {
  case CHR: return "C";
  case SET: return "S";
  }
  for (i = 0; i < nops; i++) {
    if (opeval[i] == k) {
      return opestr[i];
    }
  }
  return "NULL";
}

static int checkpclass(const char *sym, int l) {
  int i;
  for (i = 0; i < npclass; i++) {
    if (MBSCMP(sym, l, pclass[i].sym)) {
      return i;
    }
  }
  return -1;
}

static unsigned long almesccnv(const char *seq, int l) {
  int i;
  for (i = 0; i < nesc; i++) {
    if (MBSCMP(seq, l, escstr[i])) {
      return escchr[i];
    }
  }
  return 0;
}

int almexecmp(ALMEXP *e, int n, unsigned long c) {
  unsigned long *op = ALMOP(e, n);
  printf("// DEBUG:almopc(offset=%ld)\n", ALMOE(e, n));
  return almexeopc(e, op, c);
}


static int almsetcmp(const void *key, const void *val) {
  return *(unsigned long *)key - *(unsigned long *)val;
}

static int almrngcmp(const void *key, const void *val) {
  unsigned long kv = *(unsigned long *)key;  
  unsigned long lv = ((unsigned long *)val)[0];
  unsigned long hv = ((unsigned long *)val)[1];
  printf("// DEBUG:RNGCMP:lo=%lx,c=%lx,hi=%lx@%lx\n", 
   lv, kv, hv, val);
  if (kv < lv) return -1;
  else if (hv < kv) return +1;
  else return 0;
}

int almexeopc(ALMEXP *e, unsigned long *p, unsigned long c) {
  size_t i = 0;
  int neg = 0, v = 0;
  printf("// DEBUG:almopc(@%lx)\n", p);
  if (p[i] == ALMOPBLK) {
    i += 1;
    if (p[i] == ALMOPNEG) {
      i += 1;
    }
  } 
  do {
    switch (p[i]) {
    case ALMOPDOT:
      printf("// DEBUG:CMP:ALMOPDOT\n");
      v = 1;
      i += 1;
      break;
    case ALMOPCMP:
      printf("// DEBUG:CMP:%lx == %lx\n", p[i + 1], c);
      v = (p[i + 1] == c);
      i += 2;
      printf("// DEBUG:CMP=%s\n", v ? "success" : "failed");
      break;
    case ALMOPSET:
      printf("// DEBUG:CMP:ALMOPSET:%lx\n", c);
      v = (bsearch(&c, &(p[i + 2]), p[i + 1], 
       sizeof (unsigned long), almsetcmp) ? 1 : 0);
      i += (2 + p[i + 1]);
      printf("// DEBUG:CMP=%s\n", v ? "success" : "failed");
      break;
    case ALMOPRNG:
      printf("// DEBUG:CMP:ALMOPRNG:%lx:x%d\n", c, p[i + 1]);
      v = (bsearch(&c, &(p[i + 2]), p[i + 1], 
       sizeof (unsigned long) * 2, almrngcmp) ? 1 : 0);
      i += (2 + p[i + 1] * 2);
      printf("// DEBUG:CMP=%s\n", v ? "success" : "failed");
      break;
    default:
      printf("// DEBUG:CMP:UNKNOWN(%ld)\n", p[0]);
      i += 1;
      break;
    }
  } while ((p[0] == ALMOPBLK) && (p[i] != ALMOPEND) && !v);
  return v ^ neg;
}

void almopdump(FILE *fp, ALMEXP *e, int x) {
  unsigned long *p;
  char buff[8];
  size_t i;
  size_t j;
  fprintf(fp, "// S%d:", x);
  fwrite(ALMTSTR(e, ALMTN(e, x)), 1, ALMTL(e, ALMTN(e, x)), fp);
  fprintf(fp, ":offset=%lx\n", ALMOE(e, x));
  p = &(e->op[ALMOE(e, x)]);
  printf("// DEBUG:almopdump(@%lx)\n", p);
  for (i = 0; p[i] != ALMOPEND; ) {
    switch (p[i]) {
    case ALMOPBLK:
      fprintf(fp, "// ALMOPBLK\n");
      i += 1;
      break;
    case ALMOPNEG:
      fprintf(fp, "// ALMOPNEG\n");
      i += 1;
      break;
    case ALMOPDOT:
      fprintf(fp, "// ALMOPDOT\n");
      i += 1;
      break;
    case ALMOPCMP:
      if ((p[i + 1] >= 0x20) && (p[i + 1] != 0x7f)) {
        (*(e->save))(buff, p[i + 1]);
      } else {
        sprintf(buff, "\\x%02x", p[i + 1]);
      }
      fprintf(fp, "// ALMOPCMP:%s\n", buff);
      i += 2;
      break;
    case ALMOPSET:
      fprintf(fp, "// ALMOPSET[");
      for (j = 0; j < p[i + 1]; j++) {
        if ((p[i + 2 + j] >= 0x20) && (p[i + 2 + j] != 0x7f)) {
          (*(e->save))(buff, p[i + 2 + j]);
        } else {
          sprintf(buff, "\\x%02x", p[i + 2 + j]);
        }
        fprintf(fp, "%s%s", buff, (j == (p[i + 1] - 1) ? "" : ","));
      }
      fprintf(fp, "]\n");
      i += (2 + p[i + 1]);
      break;
    case ALMOPRNG:
      fprintf(fp, "// ALMOPRNG[");
      for (j = 0; j < (p[i + 1] * 2); j++) {
        if ((p[i + 2 + j] >= 0x20) && (p[i + 2 + j] != 0x7f)) {
          (*(e->save))(buff, p[i + 2 + j]);
        } else {
          sprintf(buff, "\\x%02x", p[i + 2 + j]);
        }
	if (j & 1) {
          fprintf(fp, "%s%s", buff, (j == (p[i + 1] * 2 - 1)) ? "" : ",");
	} else {
          fprintf(fp, "%s-", buff);
	}
      }
      fprintf(fp, "]\n");
      i += (2 + p[i + 1] * 2);
      break;
    case ALMOPFNC:
      fprintf(fp, "// ALMOPFNC:%s\n", pclass[p[i + 1]].sym);
      i += 2;
      break;
    default:
      fprintf(fp, "// UNKNOWN:%d\n", p[i]);
      i += 1;
      break;
    }
    if (p[0] != ALMOPBLK) break;
  }
  if (p[0] == ALMOPBLK) {
    fprintf(fp, "// ALMOPEND\n");
  }
}

void almopdumpall(FILE *fp, ALMEXP *e) {
  int i;
  for (i = 1; i < e->ns; i++) {
    almopdump(fp, e, i);
  }
}

void almerror(FILE *fp, ALMEXP *e) {
  switch (e->en) {
  case ALMLEXERR:
    fprintf(fp, "lexical error, offset %d byte\n", e->ev);
    break;
  case ALMSYNERR:
    fprintf(fp, "syntax error, near:'");
    fwrite(ALMTSTR(e, e->ev), 1, ALMTL(e, e->ev), fp);
    fprintf(fp, "'\n");
    break;
  case ALMEPSERR:
    fprintf(fp, "empty expression\n");
    break;
  case ALMSTTERR:
    fprintf(fp, "operator only\n");
    break;
  case ALMMEMERR:
    fprintf(fp, "malloc error\n");
    break;
  case ALMSTKERR:
    fprintf(fp, "parser stack overflow(current depth:%d)\n", e->ev);
    break;
  case ALMRUNERR:
    fprintf(fp, "bug (runtime?)\n");
    break;
  case ALMPCLERR:
    fprintf(fp, "unsupported POSIX class:");
    fwrite(&(e->re[(e->ev >> 16) & 0xffff]), 1, e->ev & 0xffff, fp);
    fprintf(fp, "\n");
    break;
  default:
    fprintf(fp, "unknown error:%d\n", e->en);
    break;
  }
}

void almmakedot(FILE *fp, ALMEXP *e) {
  int y, x, n;
  fprintf(fp, "digraph \"G\" {\n");
  fprintf(fp, "graph [rankdir=\"LR\", label=\"%s\"];\n", e->re);
  fprintf(fp, "node [fontname=\"MSGOTHIC.TTC\", shape=\"circle\"];\n");
  fprintf(fp, "edge [fontname=\"MSGOTHIC.TTC\"];\n");
  for (y = 0; y < e->ns; y++) {
    for (x = 0; x < e->ns; x++) {
      if ((n = ALMFOLLOWPOS(e, y, x)) != -1) {
	fprintf(fp, "S%d->S%d [label=\"", y, x);
	fwrite(ALMTSTR(e, n), 1, ALMTL(e, n), fp);
	fprintf(fp, "\"]\n");
      }
    }
  }
  for (x = 0; x < e->ns; x++) {
    if (ALMCHKBIT(e->mk, x)) {
      printf("S%d [peripheries=3];\n", x);
    } else if (ALMACC(e, x)) {
      printf("S%d [peripheries=2];\n", x);
    }
  }
  fprintf(fp, "}\n");
}

ALMEXP *almcompile(const char *s, int len, const char *charset) {
  union { ALMEXP *ep; char *cp; } up;
  static void fetch(ALMEXP *p, const char *s, int len);
  ALMEXP e;
  ALMTMP *at;
  void *ps;
  size_t n;
  int i;
  memset(&e, 0, sizeof e);
  if (strcmpi(charset, ALMENCMSJ) == 0) {
    e.clen = sjisclen;
    e.code = sjiscode;
    e.save = sjissave;
  } if (strcmpi(charset, ALMENCEUC) == 0) {
    e.clen = ujisclen;
    e.code = ujiscode;
    e.save = ujissave;
  } if (strcmpi(charset, ALMENCUTF) == 0) {
    e.clen = utf8clen;
    e.code = utf8code;
    e.save = utf8save;
  }
  fetch(&e, s, len);
  if (!e.en) {
    size_t stsz = sizeof (ALMEXP);
    size_t oesz = sizeof (unsigned long) * e.ns;
    size_t opsz = sizeof (unsigned long) * e.no;
    size_t tlsz = ALMBMLEN(e.ns);
    size_t mksz = ALMBMLEN(e.ns);
    size_t tfsz = ALMBMLEN(e.ns);
    size_t bfsz = ALMBMLEN(e.ns);
    size_t tksz = sizeof (short) * e.nt * ALMTKENT;
    size_t tnsz = sizeof (short) * e.ns;
    size_t flsz = sizeof (short) * e.ns * e.ns;
    size_t total = stsz + oesz + opsz
     + tlsz + mksz + tfsz + bfsz + tksz + tnsz + flsz 
     + (len + 1);
    if ((up.cp = malloc(total)) != NULL) {
      size_t offset = 0;
      memcpy(up.cp, &e, sizeof (ALMEXP));
      memset(&(up.cp[stsz]), 0, total - stsz); offset += stsz;
      up.ep->oe = (unsigned long *)&(up.cp[offset]); offset += oesz;
      up.ep->op = (unsigned long *)&(up.cp[offset]); offset += opsz;
      up.ep->tl = (unsigned long *)&(up.cp[offset]); offset += tlsz;
      up.ep->mk = (unsigned long *)&(up.cp[offset]); offset += mksz;
      up.ep->tf = (unsigned long *)&(up.cp[offset]); offset += tfsz;
      up.ep->bf = (unsigned long *)&(up.cp[offset]); offset += bfsz;
      up.ep->tk = (short *)&(up.cp[offset]); offset += tksz;
      up.ep->tn = (short *)&(up.cp[offset]); offset += tnsz;
      up.ep->fl = (short *)&(up.cp[offset]); offset += flsz;
      up.ep->re = &(up.cp[offset]); offset += (len + 1);
      up.ep->rl = len;
      memcpy(up.ep->re, s, len);
      up.ep->re[len] = '\0';
      memset(up.ep->fl, 0xff, flsz);
      fetch(up.ep, s, len);
      if (at = makealmtmp(up.ep, YYSTACKDEPTH)) {
        if (ps = alamoAlloc(malloc)) {
          for (i = 0; i < up.ep->nt; i++) {
            alamo(ps, ALMTV(up.ep, i), i, at);
	    if (at->cc) { /* ERROR */
	      break;
	    }
          }
          if (!at->cc) alamo(ps, EOE, 0, at);
          if (!at->cc) alamo(ps, 0, 0, at);
	  switch (at->cc) {
	  case  ALMPARSACC: /* success */
	    break;
	  case  ALMPARSNON: /* no reach */
	    up.ep->en = ALMRUNERR;
	    break;
	  case  ALMPARSSYN: /* syntax error */
	    up.ep->en = ALMSYNERR;
	    up.ep->ev = i;
	    break;
	  case  ALMPARSSTK: /* stack overflow */
	    up.ep->en = ALMSTKERR;
	    up.ep->ev = YYSTACKDEPTH;
	    break;
	  default: /* no reach */
	    printf("/* parser status error */\n");
	    break;
	  }
          alamoFree(ps, free);
        } else {
          up.ep->en = ALMMEMERR;
	}
        free(at);
      } else {
        up.ep->en = ALMMEMERR;
      }
      return up.ep;
    } else {
      return NULL;
    }
  } else {
    if ((up.cp = malloc(sizeof (ALMEXP) + len + 1)) != NULL) {
      memcpy(up.cp, &e, sizeof (ALMEXP));
      memcpy(&(up.cp[sizeof (ALMEXP)]), s, len);
      up.cp[sizeof (ALMEXP) + len] = '\0';
      up.ep->re = &(up.cp[sizeof (ALMEXP)]);
      up.ep->rl = len;
      return up.ep;
    }
  }
  return NULL;
}

#define TSTORE(p, cnt, val, off, len, num, lno) { \
  if ((p)->tk) { \
    ALMTV(p, cnt) = (val); \
    ALMTS(p, cnt) = (off); \
    ALMTL(p, cnt) = (len); \
    ALMSN(p, cnt) = (num); \
    ALMLN(p, cnt) = (lno); \
    if ((p->tn) && num) { ALMTN(p, num) = (cnt); } \
  } \
  /* \
  if (!(p)->tk) { \
     printf("TRACE[STORE:%s[%d:%d] S%d,%d]\n", \
     almopesym(val), off, len, num, cnt);  \
  } \
  */ \
  (cnt)++; \
}

#define ESTORE(p, cnt, num) { \
  /* printf("// DEBUG:S%d:off=%x\n", num, cnt); */ \
  if ((p)->oe) ALMOE(p, num) = (cnt); \
}

#define OSTORE(p, cnt, ope) { \
  /* printf("// DEBUG:cnt=%d:ope=%x\n", cnt, ope); */ \
  if ((p)->oe) { \
    (p)->op[(cnt)] = (ope); \
  } \
  (cnt)++; \
}

static void fetch(ALMEXP *e, const char *s, int len) {
  static size_t optimizecl(ALMEXP *e, size_t s, size_t l);
  unsigned long ch, cv;
  int stop = 0;
  int ln = 1; /* for multi-line expression */
  size_t nt = 0; /* space for token buff */
  size_t ip = 0; /* space for opecode */
  size_t ie = 0; /* space for POSIX class expand */
  size_t co = 0; /* column number */
  size_t cb;
  size_t im; /* marker for ip */
  int cm; /* marker for class function */
  int i; 
  int m;
  int st = S00;
  short sn = 0;
  short op;
  for (i = 0; cb = (*(e->clen))(s, i, len); i += cb) {
    ch = (*(e->code))(s, i, len);
    co += 1;
    switch (st) {
    case S00:
      switch (ch) {
      case '\\':
        m = i;
        st = S01;
        break;
      case '[':
        m = i;
	im = ip;
	OSTORE(e, ip, ALMOPBLK);
        st = S12;
        break;
      default:
	op = almopeval(s, i, cb);
	if (op == TOP) op = ((co == 1) ? TOP : 0);
	if (op == BOT) 
	 op = ((((i + cb) == len) || (s[i + cb] == '\n')) ? BOT : 0);
	if (op) {
	  if (op == DOT) {
	    sn += 1;
            TSTORE(e, nt, op, i, cb, sn, ln);
	    ESTORE(e, ip, sn);
	    OSTORE(e, ip, ALMOPDOT);
	  } else {
	    if (op == RET) {
	      co = 0;
	      ln += 1;
	    }
            TSTORE(e, nt, op, i, cb, 0, ln);
	  }
        } else {
	  sn += 1;
          TSTORE(e, nt, CHR, i, cb, sn, ln);
	  ESTORE(e, ip, sn);
	  OSTORE(e, ip, ALMOPCMP);
	  OSTORE(e, ip, ch);
	}
      }
      break;
    case S01: // backslash
      switch (ch) {
      case 'x':
        st = S02;
        break;
      case 'X':
        st = S04;
        break;
      default: // escaped 
	sn += 1;
	if (cv = almesccnv(&(s[m]), i - m + cb)) {
          TSTORE(e, nt, CHR, m, i - m + cb, sn, ln);
	  ESTORE(e, ip, sn);
	  OSTORE(e, ip, ALMOPCMP);
	  OSTORE(e, ip, cv);
	} else {
          TSTORE(e, nt, CHR, m, i - m + cb, sn, ln);
	  ESTORE(e, ip, sn);
	  OSTORE(e, ip, ALMOPCMP);
	  OSTORE(e, ip, ch);
	}
	st = S00;
        break;
      }
      break;
    case S02: case S03: // escape hex
      if (isxdigit(ch)) {
        if (st < S03) {
          st = st + 1;
        } else {
	  unsigned long v;
	  sscanf(&(s[i - 1]), "%02lx", &v);
	  sn += 1;
          TSTORE(e, nt, CHR, m, i - m + cb, sn, ln);
	  ESTORE(e, ip, sn);
	  OSTORE(e, ip, ALMOPCMP);
	  OSTORE(e, ip, v);
          st = S00;
        }
      } else {
        stop = 1;
      }
      break;
    case  S04: case  S05: case S06: case S07:
    case  S08: case  S09: case S10: case S11:
      if (isxdigit(ch)) {
        if (st < S11) {
          st = st + 1;
        } else {
	  unsigned long v;
	  sscanf(&(s[i - 7]), "%08lx", &v);
	  sn += 1;
          TSTORE(e, nt, CHR, m, i - m + cb, sn, ln);
	  ESTORE(e, ip, sn);
	  OSTORE(e, ip, ALMOPCMP);
	  OSTORE(e, ip, v);
          st = S00;
        }
      } else {
        stop = 1;
      }
      break;
    case S12:
      if ((ch == '[') 
         && !(((i + cb) < len)
	  && ((s[i + cb] == ':')
	   || (s[i + cb] == '.')
	   || (s[i + cb] == '=')))) {
        cm = i;
        st = S13;
      } else if ((ch == ']') 
       && !(i == (m + 1))
       && !((i == (m + 2)) && (s[m + 1] != '^'))) {
	sn += 1;
        TSTORE(e, nt, SET, m, i - m + cb, sn, ln);
	ESTORE(e, im, sn);
	OSTORE(e, ip, ALMOPEND);
	/* optimizecl start:im, length:(ip - im) */
	if (e->op) {
	  int sz = optimizecl(e, im, ip - im);
	  ip = (im + sz);
        } 
        st = S00;
      } else if (ch == '\\') {
        st = S20;
      } else {
        if ((i == (m + 1)) && (ch == '^')) {
	  OSTORE(e, ip, ALMOPNEG);
	} else {
	  OSTORE(e, ip, ALMOPCMP);
	  OSTORE(e, ip, ch);
	}
      }
      break;
    case S13: 
      if (ch == ':') {
        st = S14;
      } else if (ch == '=') {
        st = S16;
      } else if (ch == '.') {
        st = S18;
      } else { /* no reach */
        stop = 1;
      }
      break;
    case S14:
      if (ch == ':') {
        st = S15;
      } else {
        ;
      }
      break;
    case S15:
      if (ch == ']') {
        unsigned long v = ((cm << 16) | (i - cm + 1));
        int x = checkpclass(&(s[cm]), i - cm + 1);
	if (x < 0) {
	  e->en = ALMPCLERR;
	  e->ev = v;
	  stop = 1;
	  break;
	}
        ie += (pclass[x].sl + pclass[x].rl);
        ie += (pclass[x].sl ? 2 : 0);
        ie += (pclass[x].rl ? 2 : 0);
	OSTORE(e, ip, ALMOPFNC);
	OSTORE(e, ip, x);
        st = S12;
      } else {
        stop = 1;
      }
      break;
    case S16:
      if (ch == '=') {
        st = S17;
      } else {
        ;
      }
      break;
    case S17:
      if (ch == ']') {
        unsigned long v = ((cm << 16) | (i - cm + 1));
        int x = checkpclass(&(s[cm]), i - cm + 1);
	if (x < 0) {
	  e->en = ALMPCLERR;
	  e->ev = v;
	  stop = 1;
	  break;
	}
        ie += (pclass[x].sl + pclass[x].rl);
        ie += (pclass[x].sl ? 2 : 0);
        ie += (pclass[x].rl ? 2 : 0);
	OSTORE(e, ip, ALMOPFNC);
	OSTORE(e, ip, x);
        st = S12;
      } else {
        stop = 1;
      }
      break;
    case S18:
      if (ch == '.') {
        st = S19;
      } else {
        ;
      }
      break;
    case S19:
      if (ch == ']') {
        unsigned long v = ((cm << 16) | (i - cm + 1));
        int x = checkpclass(&(s[cm]), i - cm + 1);
	if (x < 0) {
	  e->en = ALMPCLERR;
	  e->ev = v;
	  stop = 1;
	  break;
	}
        ie += (pclass[x].sl + pclass[x].rl);
        ie += (pclass[x].sl ? 2 : 0);
        ie += (pclass[x].rl ? 2 : 0);
	OSTORE(e, ip, ALMOPFNC);
	OSTORE(e, ip, x);
        st = S12;
      } else {
        stop = 1;
      }
      break;
    case S20:
      if (ch == 'x') {
        st = S21;
      } else if (ch == 'X') {
        st = S23;
      } else {
	if (cv = almesccnv(&(s[m]), i - m + cb)) {
	  OSTORE(e, ip, ALMOPCMP);
	  OSTORE(e, ip, cv);
	} else {
	  OSTORE(e, ip, ALMOPCMP);
	  OSTORE(e, ip, ch);
	}
        st = S12;
      }
      break;
    case S21: case S22:
      if (isxdigit(ch)) {
        if (st < S22) {
          st = st + 1;
        } else {
	  unsigned long v;
	  sscanf(&(s[i - 1]), "%02lx", &v);
	  OSTORE(e, ip, ALMOPCMP);
	  OSTORE(e, ip, v);
          st = S12;
        }
      } else {
        stop = 1;
      }
      break;
    case S23: case S24: case S25: case S26:
    case S27: case S28: case S29: case S30:
      if (isxdigit(ch)) {
        if (st < S30) {
          st = st + 1;
        } else {
	  unsigned long v;
	  sscanf(&(s[i - 7]), "%08lx", &v);
	  OSTORE(e, ip, ALMOPCMP);
	  OSTORE(e, ip, v);
          st = S12;
        }
      } else {
        stop = 1;
      }
      break;
    }
    if (stop) break;
  }
  if (!e->tk) {
    e->ns = sn + 1;
    e->nt = nt;
    if (stop && e->en) {
      ;
    } else if (e->nt == 0) {
      e->en = ALMEPSERR;
    } else if (e->ns == 0) {
      e->en = ALMSTTERR;
    } else if (stop) {
      e->en = ALMLEXERR;
      e->ev = i;
    } else {
      e->en = 0;
      e->no = ip + ie;
    }
  }
  printf("// DEBUG:no=%d(%d+%d),nt=%d,ns=%d\n", 
   e->no, ip, ie, e->nt, e->ns);
}

size_t chkrngchr(ALMEXP *e, unsigned long *p, size_t l, unsigned long c) {
  size_t i;
  for (i = 0; i < l; i++) {
    if ((p[i * 2 + 0] <= c) && (c <= p[i * 2 + 1])) return i;
  }
  return l;
}

size_t chkdupchr(ALMEXP *e, unsigned long *p, size_t l, unsigned long c) {
  size_t i;
  for (i = 0; i < l; i++) {
    if (p[i] == c) return i;
  }
  return l;
}

#define DELRNG(ep, buff, pos, len) { int o; \
  for (o = (pos); o < ((len) - 1); o++) { \
    (buff)[(o * 2) + 0] = (buff)[(o * 2) + 2]; \
    (buff)[(o * 2) + 1] = (buff)[(o * 2) + 3]; \
  } \
  (len) -= 1; \
}

int delduprng(ALMEXP *e, unsigned long *p, size_t l, size_t m) {
  int i;
  int lo = p[m * 2 + 0];
  int hi = p[m * 2 + 1];
  for (i = 0; i < l; i++) {
    if (i == m) continue;
    if ((p[i * 2 + 0] <= lo) && (lo <= p[i * 2 + 1])) {
      printf("// DEBUG:delete case A:");
      CODEDUMP(stdout, e, lo); putchar('-'); CODEDUMP(stdout, e, hi);
      printf("\n");
      if (p[i * 2 + 1] < hi) p[i * 2 + 1] = hi;
      DELRNG(e, p, m, l);
      return 1;
    } else if ((p[i * 2 + 0] <= hi) && (hi <= p[i * 2 + 1])) {
      printf("// DEBUG:delete case B:");
      CODEDUMP(stdout, e, lo); putchar('-'); CODEDUMP(stdout, e, hi);
      printf("\n");
      if (lo < p[i * 2 + 0]) p[i * 2 + 0] = lo;
      DELRNG(e, p, m, l);
      return 1;
    } else if ((hi + 1) == p[i * 2 + 0]) {
      printf("// DEBUG:delete case C:");
      CODEDUMP(stdout, e, lo); putchar('-'); CODEDUMP(stdout, e, hi);
      printf("\n");
      p[i * 2 + 0] = lo;
      DELRNG(e, p, m, l);
      return 1;
    } else if ((p[i * 2 + 1] + 1) == lo) {
      printf("// DEBUG:delete case D:");
      CODEDUMP(stdout, e, lo); putchar('-'); CODEDUMP(stdout, e, hi);
      printf("\n");
      p[i * 2 + 1] = hi;
      DELRNG(e, p, m, l);
      return 1;
    }
  }
  return 0;
}

static size_t optimizerng(ALMEXP *e, unsigned long *p, size_t l) {
  size_t i;
  for (i = 0; i < l; i++) {
    while (delduprng(e, p, l, i)) l--;
  }
  return l;
}

static size_t optimizeset(ALMEXP *e, 
 unsigned long *rp, size_t rl,
 unsigned long *cp, size_t cl) {
  size_t i, j;
  for (i = 0; i < cl; ) {
    if (chkrngchr(e, rp, rl, cp[i]) != rl) {
      printf("// DEBUG:delete:"); CODEDUMP(stdout, e, cp[i]); printf("\n");
      for (j = i; j < (cl - 1); j++) { cp[j + 0] = cp[j + 1]; } 
      cl -= 1; 
    } else {
      i++;
    }
  }
  return cl;
}

static int qsortccmp(const void *l, const void *r) {
  return *(unsigned long *)l - *(unsigned long *)r;
}

static int qsortrcmp(const void *l, const void *r) {
  /* no over wrap(optimized) */
  return ((unsigned long *)l)[0] - ((unsigned long *)r)[0];
}

static size_t optimizecl(ALMEXP *e, size_t s, size_t l) {
  size_t i, cl = 0, rl = 0, fl = 0, rc;
  unsigned long *p, *c, *r, *f;
  int neg = 0;
  if ((p = malloc(sizeof (unsigned long) * l * 3)) == NULL) {
    return 0;
  }
  printf("// DEBUG:class optimize:start=%ld:len=%ld\n", s, l);
  c = p;
  r = &(p[l]);
  f = &(p[l + l]);
  for (i = s; i < (s + l); ) {
    if (e->op[i] == ALMOPBLK) {
      i += 1;
      continue;
    } else if (e->op[i] == ALMOPNEG) {
      neg = 1;
      i += 1;
      continue;
    } else if (e->op[i] == ALMOPCMP) {
      if ((e->op[i + 2] == ALMOPCMP)
       && (e->op[i + 3] == '-')
       && (e->op[i + 4] == ALMOPCMP)) {
        unsigned long lv = e->op[i + 1];
        unsigned long rv = e->op[i + 5];
	if (lv < rv) {
	  r[rl++] = lv; r[rl++] = rv;
	} else if (lv > rv) {
	  r[rl++] = rv; r[rl++] = lv;
	} else {
	  if (chkdupchr(e, c, cl, lv) == cl) {
	    c[cl++] = lv;
	  }
	}
	i += 6;
      } else {
	if (chkdupchr(e, c, cl, e->op[i + 1]) == cl) {
          c[cl++] = e->op[i + 1];
	}
	i += 2;
      }
    } else if (e->op[i] == ALMOPFNC) {
      int j;
      for (j = 0; j < pclass[e->op[i + 1]].sl; j++) {
        c[cl++] = pclass[e->op[i + 1]].set[j];
      }
      for (j = 0; j < pclass[e->op[i + 1]].rl; j++) {
        r[rl++] = pclass[e->op[i + 1]].rng[j];
      }
      i += 2;
    } else {
      break;
    }
  }
  if (fl + cl) {
    f[fl++] = ALMOPBLK;
  }
  if (neg) {
    f[fl++] = ALMOPNEG;
  }
  rc = rl / 2;
  printf("// DEBUG:rc=%d,cl=%d\n", rc, cl); 
  if (rc) {
    { int i; 
      printf("// DEBUG:RNG PASS1:");
      for (i = 0; i < (rc * 2); i++) CODEDUMP(stdout, e, r[i]); 
      printf("\n");
    }
    rc = optimizerng(e, r, rc);
    { int i; 
      printf("// DEBUG:RNG PASS2:");
      for (i = 0; i < (rc * 2); i++) CODEDUMP(stdout, e, r[i]); 
      printf("\n");
    }
    qsort(r, rc, sizeof (unsigned long) * 2, qsortrcmp);
    { int i; 
      printf("// DEBUG:RNG PASS3:");
      for (i = 0; i < (rc * 2); i++) CODEDUMP(stdout, e, r[i]); 
      printf("\n");
    }
    if (cl) {
      cl = optimizeset(e, r, rc, c, cl);
      qsort(c, cl, sizeof (unsigned long), qsortccmp);
    }
  }
  printf("// DEBUG:rc=%d,cl=%d\n", rc, cl); 
  if (cl) {
    f[fl++] = ALMOPSET;
    f[fl++] = cl;
    for (i = 0; i < cl; i++) {
      f[fl++] = c[i];
    }
  }
  if (rc) {
    f[fl++] = ALMOPRNG;
    f[fl++] = rc;
    for (i = 0; i < (rc * 2); i++) {
      f[fl++] = r[i];
    }
  }
  f[fl++] = ALMOPEND;
  for (i = 0; i < fl; i++) {
    e->op[s + i] = f[i];
  }
  free(p);
  printf("// DEBUG:class optimized size=%d\n", fl);
  return fl;
}

#if 0
int main() {
  const char *expr = "[\\x0fau-za-cb[:lower:]A-FAGcHC]*|ab\\x0bc";
  //const char *expr = "a+|[ab]+|a+ba";
  ALMEXP *e;
  if (e = almcompile(expr, strlen(expr), "SHIFT_JIS")) {
    if (!e->en) {
      almopdumpall(stdout, e);
      almmakedot(stdout, e);
    } else {
      almerror(stdout, e);
    }
    free(e);
  }
  exit(0);
}
#endif
